/*****************************************************************************

        main.cpp
        Author: Laurent de Soras, 2013

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (4 : 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include "Token.h"

#include <algorithm>
#include <list>
#include <string>
#include <vector>

#include <cassert>
#include <cstdio>



/*\\\ CLASS DEFINITIONS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



static void print_usage ()
{
	printf ("%s",
"\n"
"Usage:\n"
"\n"
"trimx264opt [options] <range_lo> <range_hi>\n"
"\n"
"Truncates x264 frame-related parameters to encode a specific segment of\n"
"the video.\n"
"\n"
"Range_lo and range_hi define a half-open interval, so range_lo is the index\n"
"of the first frame, and range_hi the index of the frame after the last frame.\n"
"You can set range_hi to 0 to indicate there is no specified end.\n"
"\n"
"Options:\n"
"\n"
"--zonein <file>  The file containing the --zones information. One zone\n"
"                 per line, no slash delimiter, same syntax as in x264.\n"
"                 [not set]\n"
"--zoneout <file> Resulting batch file setting an environment variable\n"
"                 containing the --zones information. [zones-lo-hi.bat]\n"
"--zonevar <name> Name of the environment variable where the --zones\n"
"                 parameter will be stored on the .bat output. [x264zones]\n"
"--qpin <file>    Main --qpfile input. [not set]\n"
"--qpout <file>   Resulting --qpfile file. [qpfile-lo-hi.txt]\n"
"--tcin <file>    Input timecodes, v1 or v2. [not set]\n"
"--tcout <file>   Output timecodes (same format). [timecodes-lo-hi.txt]\n"
"\n"
	);
}



static void parse_arg (int &err_index, int &err_type, std::vector <std::string> &arg_list, std::vector <std::vector <std::string> > &opt_list, int argc, char *argv [], const std::vector <std::string> &ref_opt_list)
{
	arg_list.clear ();
	opt_list.clear ();

	err_index = 0;
	err_type  = 0;
	int            index = 1;
	while (index < argc && err_index == 0)
	{
		const std::string arg = argv [index];
		if (arg.substr (0, 2) == "--")
		{
			if (std::find (ref_opt_list.begin (), ref_opt_list.end (), arg) == ref_opt_list.end ())
			{
				err_type  = 1;
				err_index = index;
			}
			else if (! (index + 1 < argc))
			{
				err_type  = 2;
				err_index = index;
			}
			else
			{
				std::vector <std::string>  opt;
				opt.push_back (arg);
				opt.push_back (argv [index + 1]);
				opt_list.push_back (opt);
				index += 2;
			}
		}
		else
		{
			arg_list.push_back (arg);
			++ index;
		}
	}
}



static bool retrieve_opt (std::string &val, const std::string &opt, const std::vector <std::vector <std::string> > &opt_list)
{
	bool           found_flag = false;
	for (std::vector <std::vector <std::string> >::const_iterator it = opt_list.begin ()
	;	it != opt_list.end () && ! found_flag
	;	++ it)
	{
		assert (it->size () >= 2);
		if ((*it) [0] == opt)
		{
			val = (*it) [1];
			found_flag = true;
		}
	}

	return (found_flag);
}



int	load_file (std::string &content, const char *filename_0)
{
	int				ret_val = 0;

	content.erase (content.begin (), content.end ());

	FILE *			f_ptr = fopen (filename_0, "r");
	if (f_ptr == 0)
	{
        printf ("*** Error: cannot open file %s\n", filename_0);
		ret_val = -1;
	}

	if (ret_val == 0)
	{
		const int		buf_len = 4095;
		char				buffer_0 [buf_len+1];
		char *			res_ptr = buffer_0;
		do
		{
			res_ptr = fgets (buffer_0, buf_len, f_ptr);
			if (res_ptr != 0)
			{
				content += buffer_0;
			}
		}
		while (res_ptr != 0);
	}

	if (f_ptr != 0)
	{
		fclose (f_ptr);
		f_ptr = 0;
	}

	return (ret_val);
}



static void	save_string_to_file (const std::string &content, const char *out_file_0, const char *content_name_0)
{
	printf ("Writing %s...", out_file_0);
	fflush (stdout);

	FILE *         f_ptr = fopen (out_file_0, "w");
	if (f_ptr == 0)
	{
		printf ("*** Error: cannot open \"%s\" for writing %s.\n", out_file_0, content_name_0);
		exit (-1);
	}

	if (fwrite (content.c_str (), 1, content.size (), f_ptr) != content.size ())
	{
		printf ("*** Error: cannot write data to \"%s\".\n", out_file_0);
		exit (-1);
	}

	fclose (f_ptr);
	f_ptr = 0;

	printf (" done.\n");
}



class Zone
{
public:
	int            _frame [2];
	std::string    _content;
};

static void process_zones (int range_lo, int /*range_hi*/, const std::string &zone_in, const std::string &zone_out, const std::string &zone_var)
{
	std::string    content;
	if (load_file (content, zone_in.c_str ()) != 0)
	{
		exit (-1);
	}

	std::vector <Zone>   zone_list;

	std::list <std::string> line_list;
	Token::process (line_list, content, '\n', LONG_MAX);

	int            line_cnt = 0;
	for (std::list <std::string>::iterator it = line_list.begin ()
	;	it != line_list.end ()
	;	++it)
	{
		std::string &  line = *it;
		line = Token::remove_white (line);
		if (! line.empty () && line [0] != '#')
		{
			std::list <std::string> token_list;
			Token::process (token_list, line, ',', LONG_MAX);
			if (token_list.size () < 3)
			{
				printf ("*** Error: malformed zone, line %d.\n", line_cnt + 1);
				exit (-1);
			}

			Zone           zone;
			int            field_cnt = 0;
			for (std::list <std::string>::iterator it = token_list.begin ()
			;	it != token_list.end ()
			;	++ it)
			{
				if (field_cnt < 2)
				{
					const char *   tok_0 = it->c_str ();
					char *         chk_0 = const_cast <char *> (tok_0);
					zone._frame [field_cnt] = strtol (tok_0, &chk_0, 10);
					if (chk_0 == tok_0)
					{
						printf ("*** Error: cannot read frame index in zone, line %d.\n", line_cnt + 1);
						exit (-1);
					}
				}
				else
				{
					zone._content += ",";
					zone._content += *it;
				}

				++ field_cnt;
			}

			zone_list.push_back (zone);
		}

		++ line_cnt;
	}

	std::string    result;
	const int      nbr_zones = int (zone_list.size ());
	for (int zone_cnt = 0; zone_cnt < nbr_zones; ++zone_cnt)
	{
		const Zone &   zone = zone_list [zone_cnt];
		if (zone._frame [1] >= range_lo)
		{
			if (! result.empty ())
			{
				result += "/";
			}
			char           tmp_0 [255+1];
			sprintf (
				tmp_0,
				"%d,%d",
				std::max (zone._frame [0] - range_lo, 0),
				zone._frame [1] - range_lo
			);
			result += tmp_0;
			result += zone._content;
		}
	}

	result = std::string ("@set ") + zone_var + "=" + result + "\n";

	save_string_to_file (result, zone_out.c_str (), "zones");
}



class QpEntry
{
public:
	int            _frame;
	std::string    _content;
};

static void process_qpfile (int range_lo, int /*range_hi*/, const std::string &qp_in, const std::string &qp_out)
{
	std::string    content;
	if (load_file (content, qp_in.c_str ()) != 0)
	{
		exit (-1);
	}

	std::vector <QpEntry>   qp_list;

	std::list <std::string> line_list;
	Token::process (line_list, content, '\n', LONG_MAX);

	int            line_cnt = 0;
	for (std::list <std::string>::iterator it = line_list.begin ()
	;	it != line_list.end ()
	;	++it)
	{
		std::string &  line = *it;
		line = Token::remove_white (line);
		if (! line.empty () && line [0] != '#')
		{
			std::list <std::string> token_list;
			Token::process (token_list, line, ' ', LONG_MAX);
			if (token_list.size () < 2)
			{
				printf ("*** Error: malformed qpfile, line %d.\n", line_cnt + 1);
				exit (-1);
			}

			QpEntry        qp_entry;
			int            field_cnt = 0;
			for (std::list <std::string>::iterator it = token_list.begin ()
			;	it != token_list.end ()
			;	++ it)
			{
				if (field_cnt == 0)
				{
					const char *   tok_0 = it->c_str ();
					char *         chk_0 = const_cast <char *> (tok_0);
					qp_entry._frame = strtol (tok_0, &chk_0, 10);
					if (chk_0 == tok_0)
					{
						printf ("*** Error: cannot read frame index in qpfile, line %d.\n", line_cnt + 1);
						exit (-1);
					}
				}
				else
				{
					qp_entry._content += " ";
					qp_entry._content += *it;
				}

				++ field_cnt;
			}

			qp_list.push_back (qp_entry);
		}

		++ line_cnt;
	}

	std::string    result;
	const int      nbr_qp = int (qp_list.size ());
	for (int qp_cnt = 0; qp_cnt < nbr_qp; ++qp_cnt)
	{
		const QpEntry &   qp_entry =qp_list [qp_cnt];
		if (qp_entry._frame >= range_lo)
		{
			char           tmp_0 [255+1];
			sprintf (tmp_0, "%d", qp_entry._frame - range_lo);
			result += tmp_0;
			result += qp_entry._content;
			result += "\n";
		}
	}

	save_string_to_file (result, qp_out.c_str (), "qps");
}



static std::string	conv_tolower (const std::string &str)
{
	std::string    result;
	for (std::string::size_type p = 0; p < str.length (); ++p)
	{
		result += tolower (str [p]);
	}

	return (result);
}



static void	process_timecodes_v1 (int range_lo, int /*range_hi*/, const std::list <std::string> &line_list, const std::string &tc_out)
{
	std::string    result;
	std::string    assume_fps;

	std::list <std::string>::const_iterator it_line = line_list.begin ();
	int            line_index = 0;
	while (it_line != line_list.end ())
	{
		const std::string line = Token::remove_white (*it_line);

		if (! line.empty () && line [0] != '#')
		{
			std::list <std::string> tok_list;

			bool           check_range_flag = true;
			if (assume_fps.empty ())
			{
				Token::process (tok_list, Token::remove_white (line), ' ', 2);
				if (tok_list.size () >= 2)
				{
					const std::string &  tok = *tok_list.begin ();
					if (conv_tolower (tok) == "assume")
					{
						assume_fps = (*it_line) + "\n";
						check_range_flag = false;
					}
				}
			}

			if (check_range_flag)
			{
				Token::process (tok_list, Token::remove_white (line), ',', LONG_MAX);
				if (tok_list.size () < 3)
				{
					printf (
						"*** Error: Invalid line %d in timecode file \"s\".\n",
						line_index + 1,
						tc_out.c_str ()
					);
					exit (-1);
				}

				char *         stop_0;
				const char *   start_0;

				std::list <std::string>::const_iterator   it_tok = tok_list.begin ();
				start_0 = it_tok->c_str ();
				++ it_tok;
				long           f1 = strtol (start_0, &stop_0, 10);
				if (stop_0 == start_0)
				{
					printf (
						"*** Error: Invalid line %d in timecode file \"s\".\n",
						line_index + 1,
						tc_out.c_str ()
					);
					exit (-1);
				}

				start_0 = it_tok->c_str ();
				++ it_tok;
				long           f2 = strtol (start_0, &stop_0, 10);
				if (stop_0 == start_0 || f1 > f2)
				{
					printf (
						"*** Error: Invalid line %d in timecode file \"s\".\n",
						line_index + 1,
						tc_out.c_str ()
					);
					exit (-1);
				}

				f1 -= range_lo;
				f2 -= range_lo;
				if (f2 >= 0)
				{
                    f1 = std::max (f1, 0L);
					char           txt_0 [127+1];
					sprintf (txt_0, "%d,%d", f1, f2);
					result += txt_0;
					while (it_tok != tok_list.end ())
					{
						result += ",";
						result += *it_tok;
						++ it_tok;
					}
					result += "\n";
				}
			}
		}

		++ it_line;
		++ line_index;
	}

	result = std::string ("# timecode format v1\n") + assume_fps + result;

	save_string_to_file (result, tc_out.c_str (), "timecodes");
}



static void	process_timecodes_v2 (int range_lo, int /*range_hi*/, const std::list <std::string> &line_list, const std::string &tc_out)
{
	std::string    result ("# timecode format v2\n");

	std::list <std::string>::const_iterator it_line = line_list.begin ();
	int            line_index = 0;
	int            frame_index = 0;
	while (it_line != line_list.end ())
	{
		const std::string line = Token::remove_white (*it_line);

		if (! line.empty () && line [0] != '#')
		{
			// We assume that the line content is valid.
			if (frame_index >= range_lo)
			{
				result += *it_line;
				result += "\n";
			}

			++ frame_index;
		}

		++ it_line;
		++ line_index;
	}

	save_string_to_file (result, tc_out.c_str (), "timecodes");
}



static void process_timecodes (int range_lo, int range_hi, const std::string &tc_in, const std::string &tc_out)
{
	std::string    content;
	if (load_file (content, tc_in.c_str ()) != 0)
	{
		exit (-1);
	}

	std::list <std::string> line_list;
	Token::process (line_list, content, '\n', LONG_MAX);

	// Identifies the format
	int				fmt = 0;	// 1 or 2, or -1 if failure
	std::list <std::string>::const_iterator it_line = line_list.begin ();
	while (it_line != line_list.end () && fmt == 0)
	{
		std::string &  line = *(line_list.begin ());
		line = Token::remove_white (line);
		line = conv_tolower (line);
		if (! line.empty ())
		{
			if (line [0] == '#')
			{
				if (line.find ("format v1") != std::string::npos)
				{
					fmt = 1;
				}
				else if (line.find ("format v2") != std::string::npos)
				{
					fmt = 2;
				}
			}
			else
			{
				fmt = -1;
			}
		}
	}

	if (fmt <= 0)
	{
		printf ("*** Error: cannot idenfity the timecode format for file \"%s\".\n");
		exit (-1);
	}
	else if (fmt == 1)
	{
		process_timecodes_v1 (range_lo, range_hi, line_list, tc_out);
	}
	else
	{
		process_timecodes_v2 (range_lo, range_hi, line_list, tc_out);
	}
}



int main (int argc, char *argv [])
{
	// Basic argument check
	std::vector <std::string>  ref_opt_list;
	ref_opt_list.push_back ("--zonein");
	ref_opt_list.push_back ("--zoneout");
	ref_opt_list.push_back ("--zonevar");
	ref_opt_list.push_back ("--qpin");
	ref_opt_list.push_back ("--qpout");
	ref_opt_list.push_back ("--tcin");
	ref_opt_list.push_back ("--tcout");

	int            err_index;
	int            err_type;
	std::vector <std::string>  arg_list;
	std::vector <std::vector <std::string> >  opt_list;
	parse_arg (err_index, err_type, arg_list, opt_list, argc, argv, ref_opt_list);
	if (err_type != 0)
	{
		if (err_type == 1)
		{
			printf ("*** Error: unknown option \"%s\".\n", argv [err_index]);
		}
		else if (err_type == 2)
		{
			printf ("*** Error: missing parameter for option \"%s\".\n", argv [err_index]);
		}
		print_usage ();
		exit (-1);
	}
	if (arg_list.size () != 2)
	{
		printf ("*** Error: wrong number of arguments.\n");
		print_usage ();
		exit (-1);
	}

	// Argument parsing
	const char *   range_lo_0 = arg_list [0].c_str ();
	char *         range_lo_chk_0 = const_cast <char *> (range_lo_0);
	int            range_lo = strtol (range_lo_0, &range_lo_chk_0, 10);
	const char *   range_hi_0 = arg_list [1].c_str ();
	char *         range_hi_chk_0 = const_cast <char *> (range_hi_0);
	int            range_hi = strtol (range_hi_0, &range_hi_chk_0, 10);
	if (   range_lo_chk_0 == range_lo_0
	    || range_hi_chk_0 == range_hi_0
	    || (range_lo >= range_hi && range_hi != 0))
	{
		printf ("*** Error: invalid range.\n");
		print_usage ();
		exit (-1);
	}

	char           buf_lohi_0 [255+1];
	sprintf (buf_lohi_0, "%06d-%06d", range_lo, range_hi);

	std::string    zone_in;
	std::string    zone_out (std::string ("zones-") + buf_lohi_0 + ".bat");
	std::string    zone_var ("x264zones");
	std::string    qp_in;
	std::string    qp_out (std::string ("qpfile-") + buf_lohi_0 + ".txt");
	std::string    tc_in;
	std::string    tc_out (std::string ("timecodes-") + buf_lohi_0 + ".txt");

	retrieve_opt (zone_in , "--zonein" , opt_list);
	retrieve_opt (zone_out, "--zoneout", opt_list);
	retrieve_opt (zone_var, "--zonevar", opt_list);
	retrieve_opt (qp_in   , "--qpin"   , opt_list);
	retrieve_opt (qp_out  , "--qpout"  , opt_list);
	retrieve_opt (tc_in   , "--tcin"   , opt_list);
	retrieve_opt (tc_out  , "--tcout"  , opt_list);

	// Zones
	if (! zone_in.empty ())
	{
		process_zones (range_lo, range_hi, zone_in, zone_out, zone_var);
	}

	if (! qp_in.empty ())
	{
		process_qpfile (range_lo, range_hi, qp_in, qp_out);
	}

	if (! tc_in.empty ())
	{
		process_timecodes (range_lo, range_hi, tc_in, tc_out);
	}

	return (0);
}



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
